/*
 * Ethernet Driver Routines
 *
 *  The TCP code uses Ethernet constants for protocol numbers and 48 bits
 *  for address.  Also, 0xFFFFFFFFFFFF is assumed to be a broadcast.
 *
 *  If you need to write a new driver, implement it at this level and use
 *  the above mentioned constants as this program's constants, not device
 *  dependant constants.
 *
 *  The packet driver code lies below this module.
 *
 *  _eth_addr      - Ethernet address of this host.
 *  _eth_brdcast - Ethernet broadcast address.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

#ifdef __HIGHC__
#include <init.h>  /* _mwenv, PL_ENV */
#endif

#include "copyrigh.h"
#include "wattcp.h"
#include "wdpmi.h"
#include "misc.h"
#include "strings.h"
#include "language.h"
#include "sock_ini.h"
#include "loopback.h"
#include "pcmulti.h"
#include "pcqueue.h"
#include "pcconfig.h"
#include "pcstat.h"
#include "pctcp.h"
#include "pcpkt.h"
#include "pcsed.h"


eth_address _eth_addr;             /* local link-layer source address  */
eth_address _eth_brdcast;          /* Link-layer broadcast address     */
WORD        _pktdevclass  = 1;     /* Ethernet = 1, SLIP = 6, PPP = 18 */
BYTE        _pktserial    = FALSE; /* using serial driver, SLIP/PPP    */
BYTE        _pkttoken     = FALSE; /* using Token-Ring driver          */
BYTE        _pktdevlevel  = 1;     /* basic unless otherwise specified */
BYTE        _pkterror     = 0;     /* error code set in pcpkt.c API    */
int         _pktretries   = 2;     /* # of retries in pkt_send loop    */
BOOL        _eth_is_init  = 0;
BOOL        _ip_recursion = 0;  /* to avoid recursion via arp_resolve() */

#if defined(USE_LOOPBACK)
static void eth_send_loopback (eth_Packet);
#endif

/*
 *  _eth_FormatPacket() places the next packet into the link-layer output
 *  buffer and uses the type field for protocol determination.
 *
 *  Note, I only maintain a single output buffer, and it gets used quickly
 *  then released.  The benefits of non-blocking systems are immense.
 */
static eth_Packet outbuf;

BYTE *_eth_formatpacket (const void *eth_dest, WORD eth_type)
{
  BYTE *ip = &outbuf.data[0];

  memset (ip, 0, sizeof(outbuf.data));

  if (_pktserial)
  {
    outbuf.type = IP_TYPE;
  }
#if 0   /* !!to-do */
  else if (_pkttoken)
  {
    memcpy (&outbuf.tr.destination, eth_dest, sizeof(tr_address));
    memcpy (&outbuf.tr.source,    &_eth_addr, sizeof(tr_address));
    outbuf.tr.type = eth_type;
  }
#endif
  else  /* DIX Ethernet encapsulation */
  {
    memcpy (&outbuf.destination, eth_dest, sizeof(eth_address));
    memcpy (&outbuf.source,    &_eth_addr, sizeof(eth_address));
    outbuf.type = eth_type;
  }
  return (ip);
}

/*
 *  _eth_send() does the actual transmission once we are complete with
 *  filling the buffer.  Do any last minute patches here, like fix the
 *  size. Send to "loopback" device if it's IP and destination matches
 *  loopback network (127.x.x.x.).
 *
 *  Don't return 'tx_len'. It may destroy e.g udp_write() if data-
 *  length < ETH_MIN-20 (40). Return length of network-layer packet.
 */
int _eth_send (WORD len)
{
  eth_Packet *tx;
  in_Header  *ip   = (in_Header*) &outbuf.data[0];
  BOOL        loop = FALSE;
  unsigned    tx_len;

  if (outbuf.type == IP_TYPE &&              /* it's an IP-packet */
      is_local_addr(intel(ip->destination))) /* to a local address */
     loop = TRUE;

  if (_pktserial)
  {
    tx = (eth_Packet*) ip;
    tx_len = len;
  }
  else
  {
    tx = (eth_Packet*) &outbuf.destination;
    tx_len = len + sizeof (eth_Header);

    if (tx_len < ETH_MIN)
        tx_len = ETH_MIN;
    if (tx_len > ETH_MAX)
        tx_len = ETH_MAX;
  }
  update_out_stat (tx);

  if (loop)
  {
#if defined(USE_LOOPBACK)
    eth_send_loopback (*tx);
#endif
    return (len);
  }

  if (pkt_send(tx, tx_len))  /* send to packet driver */
     return (len);

  if (debug_on)
     outs ("Tx failed.");
  return (0);
}


/*
 *  Initialize the Network driver interface
 *  Return error-code or 0 if okay.
 */
int _eth_init (void)
{
  if (!_eth_is_init)
  {
#if defined(__HIGHC__)
    if (_mwenv != PL_ENV)
    {
      outsnl (_LANG("Only Pharlap DOS extender supported"));
      return (WERR_ILL_DOSX);
    }

#elif defined(__WATCOM386__)   /* Watcom386 + DOS4GW style or Pharlap */
    if (dpmi_init() < 0)
       return (WERR_ILL_DOSX);

#elif (DOSX & WDOSX)           /* 32-bit Borland/MSVC + WDOSX extender */
    if (dpmi_init() < 0)
       return (WERR_ILL_DOSX);
#endif

    if (!pkt_eth_init(&_eth_addr))
       return (WERR_NO_DRIVER);  /* error message already printed */

    _eth_is_init = 1;
    memset (&_eth_brdcast, 0xFF, sizeof(_eth_brdcast));
    pkt_buf_wipe();
  }
  return (0);  /* everything okay */
}

/*
 *  Sets a new ethernet address for interface
 */
int _eth_set_addr (eth_address *eth)
{
  if (_pktserial || pkt_set_addr(eth))
  {
    memcpy (&_eth_addr, eth, sizeof(_eth_addr));
    return (1);
  }
  return (0);
}


/*
 *  _eth_free() - free an input buffer once it is no longer needed
 *  If it's a IP fragment 'pkt_free_ip()' will free the
 *  fragment buffers, buf not update the '_pkt_inf->ip_queue' queue.
 */
void _eth_free (const void *pkt, WORD type)
{
  if (type == IP_TYPE)
       pkt_free_ip (pkt);
  else pkt_free_arp (pkt);
}

/*
 *  _eth_arrived() - poll if a new packet (IP/ARP/RARP) has arrived.
 *  Sets protocol-type of packet received.
 *
 *  For ethernet-type drivers:
 *    Return value points past the ether-header into the IP/ARP/RARP
 *    protocol header. Also check if ether-broadcast.
 *
 *  For PPP/SLIP-type drivers:
 *    Return value points at the IP-packet itself.
 *    IP-protocol is assumed. Is never link-layer broadcast.
 */
BYTE *_eth_arrived (WORD *type, BOOL *brdcast)
{
  eth_Packet *eth = pkt_received();

  if (!eth)
     return (NULL);

  update_in_stat (eth);

  if (_pktserial)
  {
    *type    = IP_TYPE;
    *brdcast = FALSE;
    return ((BYTE*)eth);
  }

  /* If ip_handler() can't be reentered, only accept
   * non-IP packets
   */ 
  if (_ip_recursion && eth->type == IP_TYPE)
     return (NULL);

  *type = eth->type;
  
  if (eth->destination[0] & 1)  /* it's a link-layer broadcast */
       *brdcast = TRUE;
  else *brdcast = FALSE;

  return (&eth->data[0]);
}


#if defined(USE_LOOPBACK)
/*
 *  _eth_send_loopback() - enqueue a link-layer frame (ip only)
 *                         to the loopback device.
 *
 *  Note: this function uses call-by-value. Thus 'tx' buffer can
 *  be modified by loopback_device() and lopback handler may send
 *  using _eth_send().
 */
static void eth_send_loopback (eth_Packet tx)
{
  struct pkt_ringbuf *q;
  const  in_Header   *ip;
  int    ip_len, ip_ofs;

  if (!_pkt_inf)
  {
    IP_STAT (ips_odropped++); /* maybe this should be a input counter */
    return;
  }

  /* Call loopback handler with IP-packet
   */
  ip_ofs = pkt_get_mac_ofs();
  ip     = (in_Header*) ((char*)&tx + ip_ofs);
  ip_len = loopback_device ((in_Header*)ip);

  if (ip_len > ETH_MAX_DATA)
  {
    IP_STAT (ips_odropped++); /* maybe this should be a input counter */
    return;
  }

  q = &_pkt_inf->ip_queue;
  if (q && ip_len > 0)
  {
    eth_Packet *head;

    /* Don't let pkt_receiver() modify the queue while testing/copying.
     */
    DISABLE();
    if (pktq_in_index(q) == q->out_index)  /* queue is full, drop it */
    {
      IP_STAT (ips_odropped++);
      q->num_drop++;
      ENABLE();
      return;
    }

    head = (eth_Packet*) pktq_in_buf (q);

    /* Enqueue packet to head of input IP-queue.
     */
    if (ip_ofs)  
    {
      memcpy (head, &tx, ip_ofs);
      memcpy (&head->data[0], ip, ip_len);
    }
    else
      memcpy (head, ip, ip_len);

    /* Update queue head index
     */
    q->in_index = pktq_in_index (q);
    ENABLE();
  }
}
#endif  /* USE_LOOPBACK */


#if defined(USE_MULTICAST)   /* Jim Martin's Multicast extensions */
/*
 * _eth_join_mcast_group - joins a multicast group (at the physical layer)
 *
 * int _eth_join_mcast_group (int entry)
 * Where:
 *    entry   the entry # in the _ipmulti table
 * Returns:
 *    1       if the group was joined successfully
 *    0       if attempt failed
 */
int _eth_join_mcast_group (int entry)
{
  eth_address list [IPMULTI_SIZE];
  int         i, len, mode;
  BYTE        nextentry = 0;

  /* initialize a few things
   */
  _ipmulti[entry].active    = 0;
  _ipmulti[entry].replytime = set_timeout (0);
  _ipmulti[entry].processes = 1;

  /* Fill in the hardware address
   */
  multi_to_eth (_ipmulti[entry].ina, (BYTE*)&_ipmulti[entry].ethaddr);

  mode = _pkt_get_ip_rcv_mode();
  if (mode == -1)               /* something seems to be broken */
     return (0);

  if (mode == RM_MULTICAST2)    /* we're already in all mcast mode */
  {
    _ipmulti[entry].active = 1;
    return (1);
  }

  len = _pkt_get_multicast_list (sizeof(list), list);
  if (len == -1)
     return (0);

  /* check to see if the address is already in the list
   */
  for (i = 0; i < len/sizeof(list[0]); i++)
  {
    if (!memcmp(list[i], _ipmulti[entry].ethaddr, sizeof(list[0])))
    {
      _ipmulti[entry].active = 1;
      return (1);
    }
  }

  if (len)
     nextentry = len / sizeof(list[0]);

  memcpy (&list[nextentry], &_ipmulti[entry].ethaddr, sizeof(eth_address));
  len += sizeof (list[0]);

  if (_pkt_set_multicast_list(len,list) == -1)
  {
    if (_pkterror == NO_SPACE)     /* out of space, switch mode */
    {
      if (_pkt_set_ip_rcv_mode(RM_MULTICAST2) == -1)
         return (0);
    }
    else
      return (0);
  }

  _ipmulti[entry].active = 1;
  return (1);
}


/*
 * _eth_leave_mcast_group - leaves a multicast group (at the physical layer)
 *
 * int _eth_leave_mcast_group (int entry)
 * Where:
 *    entry   the entry # in the _ipmulti table
 * Returns:
 *    1       if the group was left successfully
 *    0       if attempt failed
 */
int _eth_leave_mcast_group (int entry)
{
  eth_address list[IPMULTI_SIZE];
  unsigned    i;
  int  len, mode;
  int  ethindex = -1;

  /* NOTE: This should be expanded to include switching back to mode 4
   * if the list of multicast addresses has shrunk sufficiently.
   *
   * First check to see if we're in mode 5. if so, mark
   * the entry as inactive and return.
   */

  mode = _pkt_get_ip_rcv_mode();
  if (mode == -1)
     return (0);

  if (mode == RM_MULTICAST2)
  {
    _ipmulti[entry].active = 0;
    return (1);
  }

  /* get the list of current multicast addresses
   */
  len = _pkt_get_multicast_list (sizeof(list), list);
  if (len < 0)
     return (0);

  /* find the apropriate entry
   */
  for (i = 0; i < len/sizeof(list[0]); i++)
      if (!memcmp(list[i],_ipmulti[entry].ethaddr,sizeof(list[0])))
         ethindex = i;

  /* if it's not in the list, just set the entry inactive and return
   */
  if (ethindex == -1)
  {
    _ipmulti[entry].active = 0;
    return (1);
  }

  /* ahh, but it _is_ in the list. So shorten the list
   * and send it back to the PD
   */
  if (ethindex+1 < (int)(len/sizeof(list[0])))
     movmem (&list[ethindex+1], &list[ethindex],
             len-(ethindex+1)*sizeof(list[0]));

  len -= sizeof (list[0]);

  if (_pkt_set_multicast_list(len,&list[0]) == -1)
  {
    if (_pkterror == NO_SPACE)     /* out of space, switch mode */
    {
      if (_pkt_set_ip_rcv_mode(RM_MULTICAST2) == -1)
         return (0);
    }
    else
      return (0);
  }

  _ipmulti[entry].active = 0;
  return (1);
}
#endif /* USE_MULTICAST */

/*
 * Turn off stack-checking because eth_release() might be called from
 * exception handler.
 */
#if defined(__HIGHC__) || defined(__WATCOMC__)
#pragma Off(check_stack)
#endif

#if defined(__BORLANDC__)
#pragma option -N-
#endif

/*
 *  _eth_release() - release the hardware driver
 */
void _eth_release (void)
{
  if (_eth_is_init)
  {
    _eth_is_init = 0;
    pkt_release();
  }
}

