/*
 *  Update and printing of ETH/IP/ICMP/IGMP/UDP/TCP input/output
 *  statistics counters.
 *
 *  G. Vanem <giva@bgnett.no> 1997
 */

#include <stdio.h>
#include <time.h>

#include "wattcp.h"
#include "strings.h"
#include "sock_ini.h"
#include "pcconfig.h"
#include "pcicmp.h"
#include "pcqueue.h"
#include "pcpkt.h"
#include "pcsed.h"
#include "pcstat.h"

int sock_stats (tcp_Socket *s, WORD *days, WORD *inactive,
                WORD *cwindow, WORD *avg,  WORD *sd)
{
  time_t now = time (NULL);

  if (s->ip_type != TCP_PROTO)
     return (0);

  if (days)     *days     = now / (3600*24) - _watt_start_day;
  if (inactive) *inactive = sock_inactive;
#if defined(USE_UDP_ONLY)
  if (cwindow)  *cwindow  = 0;
  if (avg)      *avg      = 0;
  if (sd)       *sd       = 0;
#else
  if (cwindow)  *cwindow  = s->cwindow;
  if (avg)      *avg      = s->vj_sa >> 3;
  if (sd)       *sd       = s->vj_sd >> 2;
#endif
  return (1);
}

#if !defined(USE_STATISTICS)
  void print_ip_stats  (void) {}
  void print_icmp_stats(void) {}
  void print_igmp_stats(void) {}
  void print_udp_stats (void) {}
  void print_tcp_stats (void) {}
  void print_all_stats (void) {}
  void reset_stats     (void) {}

#else  /* rest of file */

struct ethstat  ethstats;
struct ipstat   ipstats;
struct udpstat  udpstats;
struct tcpstat  tcpstats;
struct icmpstat icmpstats;
struct igmpstat igmpstats;

void reset_stats (void)
{
  memset (&ethstats, 0, sizeof(ethstats));
  memset (&ipstats,  0, sizeof(ipstats));
  memset (&udpstats, 0, sizeof(udpstats));
  memset (&tcpstats, 0, sizeof(tcpstats));
  memset (&icmpstats,0, sizeof(icmpstats));
  memset (&igmpstats,0, sizeof(igmpstats));
}

static void show_stat (const char *name, DWORD num)
{
  if (num)
    (*_printf) ("    %-12s %10lu\r\n", name, num);
}

void print_eth_stats (void)
{
  (*_printf)("ETH input stats:\r\n"
             "    IP:     drop %d, waiting %d\r\n"
             "    non-IP: drop %d, waiting %d, recv/sent %lu/%lu\r\n",
             pkt_dropped_ip(), pkt_waiting_ip(),
             pkt_dropped_arp(),pkt_waiting_arp(),
             ETH_STAT(non_ip_recv), ETH_STAT(non_ip_sent));

  (*_printf)("ETH output stats:\r\n"
             "    Tx errors    %10lu\r\n", ETH_STAT(num_tx_err));
}

void print_arp_stats (void)
{
  (*_printf)("ARP Requests:    %10lu recv\r\n"
             "    Requests:    %10lu sent\r\n"
             "    Replies:     %10lu recv\r\n"
             "    Replies:     %10lu sent\r\n",
             ETH_STAT(arp.request_recv), ETH_STAT(arp.request_sent),
             ETH_STAT(arp.reply_recv),   ETH_STAT(arp.reply_sent));
}

void print_ip_stats (void)
{
  (*_printf)("IP input stats:\r\n");
  show_stat ("total:",     IP_STAT(ips_total));
  show_stat ("badsum:",    IP_STAT(ips_badsum));
  show_stat ("badver:",    IP_STAT(ips_badvers));
  show_stat ("short:",     IP_STAT(ips_tooshort));
  show_stat ("frags:",     IP_STAT(ips_fragments));
  show_stat ("fragdrop:",  IP_STAT(ips_fragdropped));
  show_stat ("fragtime:",  IP_STAT(ips_fragtimeout));
  show_stat ("reassemble:",IP_STAT(ips_reassembled));
  show_stat ("noproto:",   IP_STAT(ips_noproto));
  show_stat ("delivered:", IP_STAT(ips_delivered));
  show_stat ("badoptions:",IP_STAT(ips_badoptions));
  show_stat ("dropped:",   IP_STAT(ips_idropped));

  (*_printf)("IP output stats:\r\n");
  show_stat ("total:",     IP_STAT(ips_rawout));
  show_stat ("noroute:",   IP_STAT(ips_noroute));
  show_stat ("frags:",     IP_STAT(ips_ofragments));
  show_stat ("dropped:",   IP_STAT(ips_odropped));
}

void print_icmp_stats (void)
{
  DWORD  total_in  = 0;
  DWORD  total_out = 0;
  int    i;
  static char *types[ICMP_MAXTYPE+1] = {
              "echoreply:", "", "",   /* 0-2 */
              "unreach:",
              "srcquench:",
              "redirect:", "", "",    /* 5-7 */
              "echorequest:",
              "router adv:", "router sol:",
              "timex:",
              "parm prob:",
              "tstamp req:", "tstamp rep:",
              "info req:",   "info rep:",
              "mask req:",   "mask rep:"
            };

  (*_printf)("ICMP input stats:\r\n");
  for (i = 0; i < ICMP_MAXTYPE; i++)
      total_in += ICMP_STAT (icps_inhist[i]);
  show_stat ("total:", total_in);

  show_stat ("badsum:", ICMP_STAT(icps_checksum));
  show_stat ("badcode:",ICMP_STAT(icps_badcode));
  show_stat ("short:",  ICMP_STAT(icps_tooshort));

  for (i = 0; i < ICMP_MAXTYPE; i++)
     show_stat (types[i], ICMP_STAT(icps_inhist[i]));

  (*_printf)("ICMP output stats:\r\n");
  for (i = 0; i < ICMP_MAXTYPE; i++)
      total_out += ICMP_STAT (icps_outhist[i]);
  show_stat ("total:", total_out);

  for (i = 0; i < ICMP_MAXTYPE; i++)
     show_stat (types[i], ICMP_STAT(icps_outhist[i]));
}

void print_igmp_stats (void)
{
  (*_printf)("IGMP input stats:\r\n");
  show_stat ("total:",        IGMP_STAT(igps_rcv_total));
  show_stat ("badsum:",       IGMP_STAT(igps_rcv_badsum));
  show_stat ("short:",        IGMP_STAT(igps_rcv_tooshort));
  show_stat ("queries:",      IGMP_STAT(igps_rcv_queries));
  show_stat ("queries (bad):",IGMP_STAT(igps_rcv_badqueries));
  show_stat ("reports:",      IGMP_STAT(igps_rcv_reports));
  show_stat ("reports (bad):",IGMP_STAT(igps_rcv_badreports));
  show_stat ("reports (grp):",IGMP_STAT(igps_rcv_ourreports));

  (*_printf)("IGMP output stats:\r\n");
  show_stat ("reports:", IGMP_STAT(igps_snd_reports));
}

void print_udp_stats (void)
{
  (*_printf)("UDP input stats:\r\n");
  show_stat ("total:",      UDP_STAT(udps_ipackets));
  show_stat ("drops:",      UDP_STAT(udps_hdrops));
  show_stat ("badsum:",     UDP_STAT(udps_badsum));
  show_stat ("no service:", UDP_STAT(udps_noport));
  show_stat ("broadcast:",  UDP_STAT(udps_noportbcast));
  show_stat ("queue full:", UDP_STAT(udps_fullsock));

  (*_printf)("UDP output stats:\r\n");
  show_stat ("total:",      UDP_STAT(udps_opackets));
}

void print_tcp_stats (void)
{
  (*_printf)("TCP input stats:\r\n");
  show_stat ("total:",       TCP_STAT(tcps_rcvtotal));
  show_stat ("drops:",       TCP_STAT(tcps_drops));
  show_stat ("badsum:",      TCP_STAT(tcps_rcvbadsum));
  show_stat ("con-attempt:", TCP_STAT(tcps_connattempt));
  show_stat ("con-drops:",   TCP_STAT(tcps_conndrops));
  show_stat ("OOB pkt:",     TCP_STAT(tcps_rcvoopack));
  show_stat ("OOB byte:",    TCP_STAT(tcps_rcvoobyte));
  show_stat ("ACK pkt:",     TCP_STAT(tcps_rcvackpack));
  show_stat ("pers-drop:",   TCP_STAT(tcps_persistdrop));

  (*_printf)("TCP output stats:\r\n");
  show_stat ("total:",       TCP_STAT(tcps_sndtotal));
  show_stat ("data pkt:",    TCP_STAT(tcps_sndpack));
  show_stat ("data byte:",   TCP_STAT(tcps_sndbyte));
  show_stat ("SYN/FIN/RST:", TCP_STAT(tcps_sndctrl));
  show_stat ("OOB pkt:",     TCP_STAT(tcps_sndurg));
  show_stat ("ACK pkt:",     TCP_STAT(tcps_sndacks));
  show_stat ("ACK dly:",     TCP_STAT(tcps_delack));
  show_stat ("retrans to:",  TCP_STAT(tcps_rexmttimeo));
  show_stat ("keepalive pr:",TCP_STAT(tcps_keepprobe));
  show_stat ("keepalive to:",TCP_STAT(tcps_keeptimeo));
}

void print_all_stats (void)
{
  print_eth_stats();
  print_arp_stats();
  print_ip_stats();
  print_icmp_stats();
#if defined(USE_MULTICAST)
  print_igmp_stats();
#endif
  print_udp_stats();
  print_tcp_stats();
}

/*
 * Called from the link-layer input routine _eth_arrived().
 */
void update_in_stat (const eth_Packet *eth)
{
  const in_Header *ip;

  if (_pktserial)   /* PD_SLIP / PD_PPP */
  {
    ip = (const in_Header*) eth;
    IP_STAT (ips_total++);
  }
  else
  {
    if (eth->type == IP_TYPE)
       IP_STAT (ips_total++);
    else
    {
      ETH_STAT (non_ip_recv++);
      if (eth->type == ARP_TYPE)
      {
        const arp_Header *arp = (const arp_Header*) &eth->data[0];

        if (arp->opcode == ARP_REQUEST)
           ETH_STAT (arp.request_recv++);
        else if (arp->opcode == ARP_REPLY)
           ETH_STAT (arp.reply_recv++);
      }
      return;
    }
    ip = (const in_Header*) &eth->data[0];
  }

  switch (ip->proto)
  {
    case ICMP_PROTO:  /* counted in pcicmp.c */
         return;

    case IGMP_PROTO:
         IGMP_STAT (igps_rcv_total++);
         return;

    case UDP_PROTO:
         UDP_STAT (udps_ipackets++);
         return;

    case TCP_PROTO:
         TCP_STAT (tcps_rcvtotal++);
         return;
  }
}

/*
 * Called from the link-layer output routine _eth_send().
 */
void update_out_stat (const eth_Packet *eth)
{
  const in_Header *ip = (const in_Header*) &eth->data[0];
  unsigned         hlen;
  DWORD            ofs;

  if (_pktserial)  /* can only be IP */
    ip = (const in_Header*) eth;

  else if (eth->type == ARP_TYPE)
  {
    const arp_Header *arp = (arp_Header*) &eth->data[0];

    ETH_STAT (non_ip_sent++);
    if (arp->opcode == ARP_REQUEST)
       ETH_STAT (arp.request_sent++);
    else if (arp->opcode == ARP_REPLY)
       ETH_STAT (arp.reply_sent++);
    return;
  }
  else if (eth->type == RARP_TYPE)
       return; /* !!to-do */

  else if (eth->type != IP_TYPE)
       return; /* impossible ?! */

  IP_STAT (ips_rawout++);   /* count raw IP (fragmented or not) */

  ofs = intel16 (ip->frag_ofs);
  ofs = (ofs & IP_OFFMASK) << 3;
   
  if (ofs)       /* don't count each fragment (only with ofs 0) */
     return;  

  hlen = in_GetHdrLen (ip);

  if (ip->proto == ICMP_PROTO)
  {
    const ICMP_PKT *icmp = (const ICMP_PKT*) ((BYTE*)ip + hlen);
    BYTE            type = icmp->unused.type;

    if (type <= ICMP_MAXTYPE)
       ICMP_STAT (icps_outhist[(int)type]++);
    return;
  }

  if (ip->proto == IGMP_PROTO)
  {
    const IGMP_packet *igmp = (const IGMP_packet*) ((BYTE*)ip + hlen);

    if (igmp->type == IGMP_REPORT)
       IGMP_STAT (igps_snd_reports++);
    return;
  }

  if (ip->proto == UDP_PROTO)
  {
    UDP_STAT (udps_opackets++);
    return;
  }

  if (ip->proto == TCP_PROTO)
  {
    tcp_Header *tcp   = (tcp_Header*) ((BYTE*)ip + hlen);
    BYTE        flags = tcp->flags & tcp_FlagMASK;
    int         tlen  = intel16 (ip->length) - hlen - (tcp->offset << 2);

    TCP_STAT (tcps_sndtotal++);

    if (tlen > 0)                 /* segments with data */
    {
      TCP_STAT (tcps_sndpack++);
      TCP_STAT (tcps_sndbyte) += tlen;
    }
    if (flags & (tcp_FlagSYN|tcp_FlagFIN|tcp_FlagRST))
       TCP_STAT (tcps_sndctrl++);
    else if (flags & tcp_FlagACK) /* ACK only */
       TCP_STAT (tcps_sndacks++);
    if (flags & tcp_FlagURG)
       TCP_STAT (tcps_sndurg++);  /* !! a bug is causing this to increment */
    return;                       /* when using DOS-PPP */
  }
}
#endif  /* USE_STATISTICS */


