#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <time.h>

#include "wattcp.h"
#include "getopt.h"
#include "gettod.h"
#include "misc.h"
#include "chksum.h"

#undef USE_FAST_CKSUM

long loops   = 20000;
long ip_size = 10000;

/*
 * Some notes: timing inchksum_fast() and inchksum() reveals that Watcom's
 * wcc386 is around 100% faster that gcc 2.95.2. And that inchksum_fast()
 * is around 2.5 times faster than inchksum() compiled with wcc386.
 * Hence using inchksum_fast() for djgpp targets gives around 450% performance
 * gain !!
 */

char * Delta_ms (clock_t dt)
{
  static char buf[10];
  double delta = 1000.0 * (double) dt / (double)CLOCKS_PER_SEC;
  sprintf (buf,"%6.1f", delta);
  return buf;
}

double timeval_diff (struct timeval *a, struct timeval *b)
{
  struct timeval d;

  d.tv_sec  = a->tv_sec  - b->tv_sec;
  d.tv_usec = a->tv_usec - b->tv_usec;
  return ((double)d.tv_sec + ((double)d.tv_usec)/1E6);
}

WORD ip_chksum (const char *ptr, int len)
{
#if 0 /* from Sniffit 0.3.5 (_packet.c) */
  register int   nleft = len;
  register WORD *w     = (WORD*)ptr;
  register int   sum   = 0;
  WORD     answer      = 0;

  while (nleft > 1)
  {
    sum += *w++;
    nleft -= 2;
  }
  if (nleft == 1)
  {
    *(BYTE*)(&answer) = *(BYTE*)w;
    sum += answer;
  }
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  return (sum);

#else
  register long sum = 0;
  register const WORD *wrd = (const WORD*) ptr;

  while (len > 1)
  {
    sum += *wrd++;
    len -= 2;
  }
  if (len > 0)
     sum += *(const BYTE*)wrd;

  while (sum >> 16)
      sum = (sum & 0xFFFF) + (sum >> 16);

  return (WORD)sum;
#endif
}

WORD ip_chksum_dummy (char *ptr, int len)
{
  ARGSUSED (ptr);
  ARGSUSED (len);
  return (0);
}

void test_checksum_speed (void)
{
  struct timeval start, now;
  int    i;
  long   j;
  char  *p;
  char  *buf = alloca (ip_size);

  for (p = buf, i = 0; i < ip_size; i += 2)
  {
    *p++ = '1';
    *p++ = '2';
  }

  /*---------------------------------------------------------------*/
  printf ("Timing inchksum()...... ");
  fflush (stdout);
  gettimeofday2 (&start, NULL);

  for (j = 0; j < loops; j++)
      inchksum (buf, ip_size);

  gettimeofday2 (&now, NULL);
  printf (" time ....%.6fs\n", timeval_diff (&now, &start));

  /*---------------------------------------------------------------*/
  printf ("Timing ip_chksum()..... ");
  fflush (stdout);
  gettimeofday2 (&start, NULL);

  for (j = 0; j < loops; j++)
      ip_chksum (buf, ip_size);

  gettimeofday2 (&now, NULL);
  printf (" time ....%.6fs\n", timeval_diff (&now, &start));

  /*---------------------------------------------------------------*/
  printf ("Timing inchksum_fast()..");
  fflush (stdout);
  gettimeofday2 (&start, NULL);

  for (j = 0; j < loops; j++)
      inchksum_fast (buf, ip_size);

  gettimeofday2 (&now, NULL);
  printf (" time ....%.6fs\n", timeval_diff (&now, &start));

  /*---------------------------------------------------------------*/
  printf ("Timing overhead........ ");
  fflush (stdout);
  gettimeofday2 (&start, NULL);

  for (j = 0; j < loops; j++)
      ip_chksum_dummy (buf, ip_size);

  gettimeofday2 (&now, NULL);
  printf (" time ....%.6fs\n", timeval_diff (&now, &start));
}

void test_checksum_correctness (void)
{
  int  i, len;
  char *buf = alloca (ip_size);

  for (i = 0; i < ip_size; i++)
      buf[i] = rand();

  for (len = 1; len < ip_size; len++)
  {
    WORD csum1 = inchksum  (buf, len);
    WORD csum2 = ip_chksum (buf, len);
    WORD csum3 = inchksum_fast (buf, len);
    printf ("inchksum = %04X, ip_chksum = %04X, inchksum_fast = %04X %c %c\n",
            csum1, csum2, csum3,
            (csum1 != csum2 ? '!' : ' '),
            (csum1 != csum3 ? '!' : ' '));
  }
}

void Handler (int sig)
{
  exit (0);
}

void Usage (void)
{
  puts ("Usage: tst [-l loops] [-i ip-size] <-s | -c> \n"
        "  -s : test inchksum() speed\n"
        "  -c : test inchksum() correctness\n");
  printf ("default loops = %ld, ip-size %ld\n"
          "results from inchksum() only significant when `USE_FAST_CKSUM'"
          " is not defined\n",
          loops, ip_size);
  exit (-1);
}

int main (int argc, char **argv)
{
  int cflag = 0;
  int sflag = 0;
  int ch;

  if (argc < 2)
     Usage();

  while ((ch = getopt(argc, argv, "l:i:sc")) != EOF)
    switch (ch)
    {
      case 'l':
           loops = atol (optarg);
           break;
      case 'i':
           ip_size = atol (optarg);
           break;
      case 's':
           sflag = 1;
           break;
      case 'c':
           cflag = 1;
           break;
      default:
           Usage();
           break;
    }

  srand (time(NULL));
  init_misc();
  signal (SIGINT, Handler);

  if (sflag)
     test_checksum_speed();
  if (cflag)
     test_checksum_correctness();
  return (0);
}

