/*
 *  Module for handling timers.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "wattcp.h"
#include "misc.h"
#include "ioport.h"

int has_8254  = 0; /* Do we have a working 8254 timer chip? */
int has_rdtsc = 0; /* Do we have a CPU with RDTSC instruction? */

/*
 * System clock at BIOS location 40:6C (dword). Counts upwards.
 */
#define BIOS_CLK 0x46C

static DWORD date    = 0;
static DWORD date_ms = 0;

#if defined(HAVE_UINT64) && defined(NOT_YET)
static uint64 clocks_per_msec;
#endif


#if defined(USE_NEW_TIMERS)   /* not yet */
/*
 * New logic for handling timers. These are only for "internal" use.
 * I.e. for use by tcp-state machine etc.
 *
 * These functions tries to bypass the limitation of current design
 * where a time-change (in DOS or application) affects timeout for
 * running timers. We keep list of timers (max MAX_TIMERS) sorted on
 * expirary. The list is checked from tcp_tick().
 *
 * Modified versions from Phil Karn's KA9Q.
 */

#define MAX_TIMERS 10

struct timer {
       DWORD          c_time;     /* absolute expiry time */
       int            c_handle;   /* return handle, 0..MAX_TIMERS-1 */
       int            c_alloc;    /* 1: allocated timer, 0: free */
       int            c_expired;  /* 1: expired, 0: not expired yet */
       struct timer  *c_next;     /* link to next allocated timer */
     };

static struct timer *callout = NULL;      /* timer callout list */
static struct timer  timers [MAX_TIMERS]; /* pre-allocated timers */

#endif

#if defined(NOT_YET)   /* use RDTSC for precise timing */
#if defined(__DJGPP__)
static __inline__ void get_cpuid (DWORD val, DWORD *eax, DWORD *ebx, DWORD *ecx, DWORD *edx)
{
  __asm__ __volatile__ (
         ".byte 0x0f,0xa2"   /* cpuid opcode */
         : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
         : "0" (val));
}

static __inline__ void get_rdtsc (uint64 *tsc)
{
  __asm__ __volatile__ (
         ".byte 0x0f, 0x31"  /* rdtsc opcode */
         : "=a" (tsc[0]), "=d" (tsc[1]) : );
}

#elif defined(__WATCOM386__)
extern void get_cpuid (DWORD val, DWORD *eax, DWORD *ebx, DWORD *ecx, DWORD *edx);
#pragma aux get_cpuid =     \
        ".586"              \
        "push eax"          \
        "push ebx"          \
        "push ecx"          \
        "push edx"          \
        "mov eax, esi"      \
        "cpuid"             \
        "mov esi, [esp]"    \
        "mov [esi], edx"    \
        "mov esi, [esp+4]"  \
        "mov [esi], ecx"    \
        "mov esi, [esp+8]"  \
        "mov [esi], ebx"    \
        "mov esi, [esp+12]" \
        "mov [esi], eax"    \
        "add esp, 16"       \
        parm [esi] [eax] [ebx] [ecx] [edx] \
        modify [esi eax ebx ecx edx];

extern void get_rdtsc (uint64 *tsc);
#pragma aux get_rdtsc =
#endif
#endif   /* NOT_YET */

/*
 * init_timers() - setup timer handling.
 * Called from init_misc()
 */
void init_timers (void)
{
#if (DOSX)         /* problems using 64-bit types in small/large models */
  has_8254 = 1;    /* !! to-do: check if 8254 PIT is really working */
#endif

#if defined(NOT_YET)
#if defined(__DJGPP__) || defined(__WATCOM386__)
  if (cpu_type >= 486) /* All 486+ have CPUID */
  {
    CheckCpuType();    /* further CPU tests */
    if (get_cpu_speed(&clocks_per_msec))
       has_rdtsc = 1;
  }
#endif
#endif

  chk_timeout (0); /* init 'date' variables */

#ifdef USE_NEW_TIMERS
  memset (&timers, 0, sizeof(timers));
#endif
}

/*
 * When using Watt-32 together with programs that reprogram the PIT
 * (8254 timer chip), the millisec_clock() function may return wrong
 * values. In such cases it's best to use old-style 55ms timers.
 * Using Watt-32 with Allegro graphics library is one such case where
 * application program should call `hires_timer(0)'.
 *
 * !!to-do: A viable solution is to use the RDTSC instruction
 *          on Pentiums when that is detected (has_rdtsc=1).
 *          64-bit timestamps will wrap after ~830 years on a
 *          700MHz PC !!
 */
void hires_timer (int on)
{
  has_8254 = on;
  chk_timeout (0);
}

/*
 * Compare two timers with expirary at 't1' and 't2'.
 *    returns -1 if t1 expires before t2
 *             0 if t1 == t2
 *            +1 if t1 expires after t2
 *
 * FIX-ME: This logic fails when timers approaches ULONG_MAX
 *         after approx. 50 days.
 */
int cmp_timers (DWORD t1, DWORD t2)
{
  if (t1 == t2)
     return (0);
  return (t1 < t2 ? -1 : +1);
}

/*
 * The following 2 routines are modified versions from KA9Q NOS
 * by Phil Karn.
 *
 * clockbits() - Read low order bits of timer 0 (the TOD clock)
 * This works only for the 8254 chips used in ATs and 386s.
 *
 * The timer 0 runs in mode 3 (square wave mode), counting down
 * by twos, twice for each cycle. So it is necessary to read back the
 * OUTPUT pin to see which half of the cycle we're in. I.e., the OUTPUT
 * pin forms the most significant bit of the count. Unfortunately,
 * the 8253 in the PC/XT lacks a command to read the OUTPUT pin...
 */
static DWORD clockbits (void)
{
  VOLATILE DWORD stat, count;

  do
  {
    _outportb (0x43, 0xC2);        /* Read Back Command (timer 0, 8254 only) */
    stat   = _inportb (0x40);      /* get status of timer 0 */
    count  = _inportb (0x40);      /* LSB of count */
    count |= _inportb (0x40) << 8; /* MSB of count */
  }
  while (stat & 0x40);             /* reread if NULL COUNT bit set */

  stat = (stat & 0x80) << 8;       /* Shift OUTPUT to msb of 16-bit word */
  count >>= 1;                     /* count /= 2 */
  if (count == 0)
     return (stat ^ 0x8000UL);     /* return complement of OUTPUT bit */
  return (count | stat);           /* Combine OUTPUT with counter */
}

/*
 * Return hardware time-of-day in milliseconds. Resolution is improved
 * beyond 55 ms (the clock tick interval) by reading back the instantaneous
 * 8254 counter value and combining it with the clock tick counter.
 *
 * Reading the 8254 is a bit tricky since a tick could occur asynchronously
 * between the two reads in clockbits(). The tick counter is examined before
 * and after the hardware counter is read. If the tick counter changes, try
 * again.
 * Note: the hardware counter (lo) counts down from 65536 at a rate of
 *       1.19318 MHz (4.77/4). System tick count (hi) counts up.
 *       Fraction 11/13125 gives us a milli-sec rate count.
 */
static DWORD millisec_clock (void)
{
  DWORD hi, lo;
  do  
  {
    hi = peekl (0, BIOS_CLK);
    lo = clockbits();
  }
  while (hi != peekl(0, BIOS_CLK));   /* tick count changed, try again */

#ifdef HAVE_UINT64
  {
    uint64 x = ((uint64)hi << 16) - lo;
    return (x * 11UL / 13125UL);
  }
#else /* Emulating this would be slow. We'd better have a math-processor */
  {
    double x = ldexp ((double)hi, 16) - (double)lo;  /* x = hi*2^16 - lo */
    return (DWORD) (x * 4.0 / 4770.0);
  }
#endif
}

#if defined(HAVE_UINT64)
/*
 * Return hardware time-of-day in microseconds.
 */
uint64 microsec_clock (void)
{
  DWORD  hi, lo;
  uint64 x;
  do
  {
    hi = peekl (0, BIOS_CLK);
    lo = clockbits();
  }
  while (hi != peekl(0, BIOS_CLK));   /* tick count changed, try again */

  x = ((uint64)hi << 16) - lo;
  return (x * 4000000 / 4770000);
}
#endif


/*
 * Return time for when given timeout (msec) expires
 */
DWORD set_timeout (DWORD msec)
{
#if (DOSX) && 0  /* !!to-do */
  if (has_rdtsc)
     return millisec_cpu_clock());
#endif

  if (!has_8254)
  {
    DWORD ticks;

    if (msec > 0 && msec < 55)
         ticks = 1;
    else ticks = msec / 55UL;
    return (ticks + date + (DWORD)peekl (0,BIOS_CLK));
  }
  return (msec + date_ms + millisec_clock());
}

/*
 * Check if milli-sec value has expired:
 *   returns 1 timer expired
 *           0 not expired (or value not set)
 */
int chk_timeout (DWORD value)
{
  static char  oldHour = -1;
  static DWORD now_ms;
  static DWORD now;
  char   hour;

  now  = (DWORD) peekl (0, BIOS_CLK);
  hour = (char)(now >> 16U);

  if (hour != oldHour)
  {
    if (hour < oldHour)            /* midnight passed */
    {
      #define CORR 290UL           /* experimental correction */
      date    += 1572750UL + CORR; /* Update date (0x1800B0UL) */
      date_ms += 3600*24*1000UL;   /* ~2^26 */
    }
    oldHour = hour;
  }

#if (DOSX) && 0    /* !!to-do */
  if (has_rdtsc)   /* needs 64-bit values */
  {
    if (value == 0L)
       return (0);
    now_ms = millisec_cpu_clock());
    return (now_ms >= value);
  }
#endif


#if 0
  if (has_8254)
     now_ms = millisec_clock();

  if (value == 0L)           /* timer not initialised */
     return (0);             /* (or stopped) */
#else

  if (value == 0L)           /* timer not initialised */
     return (0);             /* (or stopped) */

  if (has_8254)
     now_ms = millisec_clock();
#endif

  if (!has_8254)
  {
    now += date;             /* date extend current time */
    return (now >= value);   /* return true if expired */
  }
  return (now_ms + date_ms >= value);
}

/*
 * Must be called by user right before or after a time change occurs.
 */
int set_timediff (long msec)
{
  date_ms -= msec;
  date    -= msec/55;
  return (0);
}

/*
 * Return time difference between timers 'now' and 't'.
 * Check for day rollover. Max timer distance is 24 hours.
 */
DWORD get_timediff (DWORD now, DWORD tim)
{
  DWORD dt;

  chk_timeout (0);

  if (now < tim)           /* passed midnight */
       dt = tim - now;
  else dt = now - tim;

#if 1
  if (has_8254)
       dt = dt % (3600*24*1000UL);
  else dt = dt % (1572750UL + CORR);
#else
  if (has_8254)
       dt %= date_ms;
  else dt %= date;
#endif
  return (dt);
}


#if defined(USE_DEBUG)
/*
 * time_str() - return string "x.xx" for timeout value.
 */
const char *time_str (DWORD val)
{
  static char buf[30];

  if (has_8254 || has_rdtsc)
       sprintf (buf, "%.3f", (double)val/1000.0);
  else sprintf (buf, "%.2f", (double)val/18.2);
  return (buf);
}
#endif /* USE_DEBUG */
 

#ifdef USE_NEW_TIMERS    /* not yet */
/*
 * Allocate and set timer to expire in specified # of milli-sec.
 * Return allocated timer handle.
 */
int timer_set (DWORD msec)
{
  struct timer *newp, *p, **pp;
  int    handle;

  for (handle = 0; handle < DIM(timers); handle++)
      if (timers[handle].c_alloc == FALSE)
         break;

  if (handle == DIM(timers))
  {
    (*_printf) ("timer: No free timers\r\n");
    exit (-1);
  }

  newp = timers + handle;
  newp->c_time = millisec_clock() + msec;

  /* Find correct place and link it in.
   */
  for (pp = &callout; (p = *pp) != NULL; pp = &p->c_next)
      if (newp->c_time < p->c_time)
         break;

  newp->c_next   = p;
  newp->c_handle = handle;
  *pp = newp;
  return (handle);
}

/*
 * Clear allocated timer handle.
 */
int timer_clr (int handle)
{
  struct timer **copp, *freep;

  /* Find matching timeout and remove it from the callout list.
   */
  for (copp = &callout; (freep = *copp) != NULL; copp = &freep->c_next)
  {
    if (freep->c_handle == handle)
    {
      *copp = freep->c_next;
      freep->c_alloc = FALSE;
      return (handle);
    }
  }
  return (-1);
}

/*
 * Check allocated timer handle for expiry.
 */
int timer_chk (int handle)
{
  return (handle >= 0            &&
          handle < DIM(timers)   &&
          timers[handle].c_alloc &&
          timers[handle].c_expired);
}

/*
 * Test first timer in callout list for expiry.
 */
int calltimeout (void)
{
  struct timer *p;
  DWORD  now = millisec_clock();
  int    num = 0;

  while (callout)
  {
    p = callout; 
    if (p->c_time > now)  /* no, it's not time yet */
       break;            

    callout = p->c_next;
    p->c_expired = TRUE;
    num++;
  }
  return (num);
}
#endif  /* USE_NEW_TIMERS */
